#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>
#include <glib.h>
#include "if_con_info.h"
#include "config.h"


gint if_get_con_info_from_chat_report(
	const gchar *chat_report_path,
	const gchar *if_name,
	IFConInfo *con_info_buf
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)
#define STRPFX(s,p)     ((((s) != NULL) && ((p) != NULL)) ? \
			(!strncmp((s), (p), strlen(p))) : FALSE)
#define ISBLANK(c)		(((c) == ' ') || ((c) == '\t'))


/*
 *	Reads and obtains connection information from the chat(8)
 *	report file.
 *
 *	Returns 0 on success or non-zero on error. If the report file
 *	exists but no connection information was found in it then -1
 *	will be returned with errno set to ENODATA.
 */
gint if_get_con_info_from_chat_report(
	const gchar *chat_report_path,
	const gchar *if_name,
	IFConInfo *con_info_buf
)
{
	FILE *fp;
	struct stat stat_buf;
	gint ninstances;
	const gchar *s;
	gchar buf[1024];

	if(con_info_buf == NULL)
	{
		errno = EINVAL;
		return(-2);
	}

	/* Reset the return values */
	(void)memset(
		con_info_buf,
		0x00,
		sizeof(IFConInfo)
	);

	if(chat_report_path == NULL)
	{
		errno = EINVAL;
		return(-2);
	}

	/* Open the chat(8) report file for reading */
	fp = fopen((const char *)chat_report_path, "rb");
	if(fp == NULL)
		return(-1);

	/* Get the connection start time from the report file's last
	 * modified time
	 */
	if(!fstat(fileno(fp), &stat_buf))
	{
		con_info_buf->start_time_sec = stat_buf.st_mtime;
	}

	/* Begin reading the chat(8) report file and get all connection
	 * information with subsequent connection information
	 * overwriting the previous until the final connection
	 * information is obtained so that we obtain the latest
	 * information
	 */
	ninstances = 0;
	while(fgets(buf, sizeof(buf), fp) != NULL)
	{
		s = buf;
		while(ISBLANK(*s))
			s++;

		/* Skip any line that does not contain the "chat:"
		 * prefix
		 */
		if(!STRPFX(s, "chat:"))
			continue;

		/* Look for the "CONNECT" string on this line */
		s = (const gchar *)strstr((const char *)s, "CONNECT");
		if(s != NULL)
		{
			gint len;
			const gchar *s_next;

			/* Seek past the "CONNECT" and any spaces to the
			 * start of the connection flags
			 */
			s += strlen("CONNECT");
			while(ISBLANK(*s))
				s++;

			/* Seek to the first flag which is assumed to
			 * be the connection speed so stop at the
			 * first numeric character
			 */
			while((*s != '\0') && !isdigit((int)*s))
				s++;

			/* Seek to the end of this value */
			s_next = s;
			while((*s_next != '\0') && (*s_next != '/') &&
			      (*s_next != '\n')
			)
				s_next++;

			/* Get the connection speed as the first flag */
			con_info_buf->speed_bits_per_sec = (gulong)atol((const char *)s);

			s = s_next;
			if(*s == '/')
				s++;
			while(ISBLANK(*s))
				s++;


			/* Get the rest of the string as the flags */
			s_next = s;
			while((*s_next != '\0') && (*s_next != '\n'))
				s_next++;
			len = s_next - s;
			if(len > 0)
			{
				const gint	buf_len = sizeof(con_info_buf->flags_string),
						len_to_copy = MIN(
					len,
					buf_len - 1
				);
				gchar *buf = con_info_buf->flags_string;
				(void)strncpy(
					buf,
					s,
					(size_t)(len_to_copy * sizeof(gchar))
				);
				buf[len_to_copy] = '\0';
			}
/*			s = s_next; */

			/* Count this line as one instance of the
			 * connection information obtained and keep
			 * reading to obtain subsequent lines since the
			 * last line probably contains the most recent
			 * connection
			 */
			ninstances++;
		}
	}

	/* Close the file */
	(void)fclose(fp);

	if(ninstances >= 1)
	{
		return(0);
	}
	else
	{
		errno = ENODATA;
		return(-1);
	}
}
