#include <gtk/gtk.h>
#include "guiutils.h"
#include "guirgbimg.h"
#include "ifm_win_gauge.h"
#include "config.h"


/*
 *	Flags:
 */
typedef enum {
	IFM_WIN_GAUGE_REALIZED	= (1 << 0)
} IFMWinGaugeFlags;


/* Callbacks */
static void ifm_win_gauge_data_destroy_cb(gpointer data);
static void ifm_win_gauge_realize_cb(GtkWidget *widget, gpointer data);
static gint ifm_win_gauge_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

/* IFMWinGauge */
IFMWinGauge *ifm_win_gauge_new(
	GtkWidget *parent,
	const gchar *widget_name,
	const GtkOrientation orientation,
	const gchar *label_text, const gint label_width,
	const IFMWinGaugeStyle gauge_style,
	const gint gauge_length,
	const gint nticks
);
void ifm_win_gauge_delete(IFMWinGauge *gauge);

/* Resize */
void ifm_win_gauge_resize(
	IFMWinGauge *gauge,
	const gint width, const gint height
);

/* Draw */
void ifm_win_gauge_draw(IFMWinGauge *gauge);
static void ifm_win_gauge_draw_continuous(IFMWinGauge *gauge);
static void ifm_win_gauge_draw_discrete(IFMWinGauge *gauge);
static void ifm_win_gauge_draw_led(IFMWinGauge *gauge);
static void ifm_win_gauge_draw_render_base(
	guint8 *tar_rgba,
	const gint tar_width, const gint tar_height,
	const gint tar_bpl,
	const guint8 r,
	const guint8 g,
	const guint8 b
);
static void ifm_win_gauge_draw_render_tick(
	IFMWinGauge *gauge,
	const gint tar_width, const gint tar_height,
	const gint tar_bpl,
	guint8 *tar_rgba,
	const gint src_width, const gint src_height,
	const gint src_bpl,
	const guint8 *src_gray,
	const gint x, const gint y,
	const gfloat red_coeff, const gfloat green_coeff, const gfloat blue_coeff
);
void ifm_win_gauge_queue_draw(IFMWinGauge *gauge);

/* IFMWinGaugeStyle */
IFMWinGaugeStyle ifm_win_gauge_get_gauge_style(IFMWinGauge *gauge);
void ifm_win_gauge_set_gauge_style(
	IFMWinGauge *gauge,
	const IFMWinGaugeStyle gauge_style
);

/* Get/Set Value */
gfloat ifm_win_gauge_get_value(IFMWinGauge *gauge); 
void ifm_win_gauge_set_value(
	IFMWinGauge *gauge,
	const gfloat value
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define IFM_WIN_GAUGE_KEY		"IFMWinGauge"


/*
 *	IFMWinGauge:
 */
struct _IFMWinGauge {
	GtkWidget	*toplevel;
	gint		freeze_count;
	IFMWinGaugeFlags	flags;
	GtkOrientation	orientation;
	IFMWinGaugeStyle	gauge_style;
	GtkWidget	*label,
			*drawing_area;
	gint		width,			/* Size of drawing_buf */
			height;
	guint8		*drawing_area_rgba;	/* Buffer for drawing_area
						 * in RGBA format */
	gint		nticks;			/* For IFM_WIN_GAUGE_STYLE_LEDS
						 * or IFM_WIN_GAUGE_STYLE_GRADIENT_LEDSS */
	gint		tick_width,
			tick_height;
	guint8		*tick_gray;		/* Buffer describing one "tick"
						 * image in grayscale format */
	gfloat		value;			/* Current value from 0.0 to
						 * 1.0 */
};


/*
 *	Toplevel GtkObject data "destroy" signal callback.
 */
static void ifm_win_gauge_data_destroy_cb(gpointer data)
{
	IFMWinGauge *gauge = IFM_WIN_GAUGE(data);
	if(gauge == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_gauge_data_destroy_cb(): gauge=%p\n", gauge);
#endif

	/* Handle this signal even if already frozen */

	gauge->freeze_count++;

/*	gauge->toplevel = NULL; */
	g_free(gauge->drawing_area_rgba);
	g_free(gauge->tick_gray);

	gauge->freeze_count--;

	g_free(gauge);
}

/*
 *	GtkDrawingArea "realize" signal callback.
 */
static void ifm_win_gauge_realize_cb(GtkWidget *widget, gpointer data)
{
	GdkWindow *window;
	GtkStyle *style;
	IFMWinGauge *gauge = IFM_WIN_GAUGE(data);
	if((widget == NULL) || (gauge == NULL))
		return;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_gauge_realize_cb(): gauge=%p\n", gauge);
#endif

	window = widget->window;
	if((window == NULL) || (gauge->flags & IFM_WIN_GAUGE_REALIZED))
		return;

	/* Handle this signal even if already frozen */

	gauge->freeze_count++;

	/* Set the style's base color as the GdkWindow's background
	 * color
	 */
	style = gtk_widget_get_style(widget);
	if(style != NULL)
		gdk_window_set_background(
			window,
			&style->base[GTK_STATE_NORMAL]
		);

	/* Mark the IFMWinGauge as realized */
	gauge->flags |= IFM_WIN_GAUGE_REALIZED;

	gauge->freeze_count--;
}

/*
 *	GtkDrawingArea any event signal callback.
 */
static gint ifm_win_gauge_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventConfigure *configure;
	GdkEventExpose *expose;
	GdkEventButton *button;
	IFMWinGauge *gauge = IFM_WIN_GAUGE(data);
	if((widget == NULL) || (event == NULL) || (gauge == NULL))
		return(status);

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
		configure = (GdkEventConfigure *)event;
		ifm_win_gauge_resize(
			gauge,
			configure->width,
			configure->height
		);
		status = TRUE;
		break;

	  case GDK_EXPOSE:
		expose = (GdkEventExpose *)event;
		ifm_win_gauge_draw(gauge);
		status = TRUE;
		break;

	  case GDK_BUTTON_PRESS:
		button = (GdkEventButton *)event;
		if(button->button == GDK_BUTTON3)
		{

		}
		break;
	}
	return(status);
}


/*
 *	Creates a new IFMWinGauge.
 */
IFMWinGauge *ifm_win_gauge_new(
	GtkWidget *parent,
	const gchar *widget_name,
	const GtkOrientation orientation,
	const gchar *label_text, const gint label_width,
	const IFMWinGaugeStyle gauge_style,
	const gint gauge_length,
	const gint nticks
)
{
	const gboolean	vertical = (orientation == GTK_ORIENTATION_VERTICAL) ?
				TRUE : FALSE;
	const gint border_minor = 2;
	GtkWidget	*w,
			*parent2, *parent3;
	IFMWinGauge *gauge = IFM_WIN_GAUGE(g_malloc0(sizeof(IFMWinGauge)));
	if(gauge == NULL)
		return(NULL);

	gauge->freeze_count++;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_gauge_new(): gauge=%p: label_text=\"%s\"\n", gauge, label_text);
#endif


	gauge->orientation = orientation;
	gauge->gauge_style = gauge_style;
	if(vertical)
	{
		gauge->width = 8;
		gauge->height = MAX(gauge_length, 1);
	}
	else
	{
		gauge->width = MAX(gauge_length, 1);
		gauge->height = 8;
	}
	gauge->nticks = CLIP(
		nticks,
		1, IFM_WIN_MAX_GAUGE_TICKS
	);

	/* Toplevel GtkBox */
	gauge->toplevel = w = vertical ?
		gtk_vbox_new(FALSE, border_minor) :
		gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_object_set_data_full(
		GTK_OBJECT(w), IFM_WIN_GAUGE_KEY,
		gauge, ifm_win_gauge_data_destroy_cb
	);
	gtk_widget_show(w);
	parent2 = w;

	/* GtkAlighment to align the GtkLabel */
	if(vertical)
	{
		w = gtk_alignment_new(0.5f, 1.0f, 0.0f, 0.0f);
		gtk_widget_set_usize(
			w,
			-1,
			label_width
		);
	}
	else
	{
		w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
		gtk_widget_set_usize(
			w,
			label_width,
			-1
		);
	}
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* GtkLabel */
	gauge->label = w = gtk_label_new(label_text);
	gtk_label_set_justify(
		GTK_LABEL(w),
		vertical ? GTK_JUSTIFY_CENTER : GTK_JUSTIFY_RIGHT
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);

	/* GtkBox to align the GtkDrawingArea */
	w = vertical ?
		gtk_hbox_new(TRUE, 0) :
		gtk_vbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* GtkDrawingArea */
	gauge->drawing_area = w = gtk_drawing_area_new();
	if(widget_name != NULL)
		gtk_widget_set_name(w, widget_name);
	gtk_widget_set_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "realize",
		GTK_SIGNAL_FUNC(ifm_win_gauge_realize_cb), gauge
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(ifm_win_gauge_event_cb), gauge
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(ifm_win_gauge_event_cb), gauge
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(ifm_win_gauge_event_cb), gauge
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(ifm_win_gauge_event_cb), gauge
	);
	gtk_widget_set_usize(
		w,
		gauge->width, gauge->height
	);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);

	gauge->freeze_count--;

	return(gauge);
}

/*
 *	Deletes the IFMWinGauge.
 */
void ifm_win_gauge_delete(IFMWinGauge *gauge)
{
	if(gauge == NULL)
		return;

#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_gauge_delete(): gauge=%p\n", gauge);
#endif
	/* Call ifm_win_gauge_data_destroy_cb() to delete the rest of the
	 * IFMWinGauge
	 */
	GTK_WIDGET_DESTROY(gauge->toplevel);
}


/*
 *	Resizes the IFMWinGauge.
 */
void ifm_win_gauge_resize(
	IFMWinGauge *gauge,
	const gint width, const gint height
)
{
#if (IFM_DEBUG_LEVEL >= 2)
g_print("ifm_win_gauge_resize(): gauge=%p: Size %ix%i.\n", gauge, width, height);
#endif

	gauge->width = width;
	gauge->height = height;

	/* Reallocate the IFMWinGauge GtkDrawingArea's RGBA image data
	 * to the new size
	 */
	g_free(gauge->drawing_area_rgba);
	if((gauge->width > 0) && (gauge->height > 0))
		gauge->drawing_area_rgba = (guint8 *)g_malloc(
			gauge->width * gauge->height * 4 * sizeof(guint8)
		);
	else
		gauge->drawing_area_rgba = NULL;


	/* Resize the tick by recreating and resizing the tick from
	 * the source tick template
	 */
	if(gauge->nticks > 0)
	{
		const gint	nticks = gauge->nticks,
				tick_width = 16,
				tick_height = 16;
		const guint8 src_tick_template_gray[] = {
0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00,
0x00, 0x40, 0x80, 0x80,  0x80, 0x80, 0x80, 0x80,  0x80, 0x80, 0x70, 0x50,  0x20, 0x00, 0x00, 0x00,
0x00, 0x80, 0xA0, 0xA0,  0xA0, 0xA0, 0xA0, 0xA0,  0xA0, 0xA0, 0xA0, 0x90,  0x70, 0x30, 0x00, 0x00,
0x00, 0xA0, 0xF0, 0xFA,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFA, 0xF0,  0xE0, 0xD0, 0xB0, 0x50,

0x00, 0xA0, 0xF0, 0xFA,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFA, 0xF0,  0xE0, 0xD0, 0xA0, 0x50,
0x00, 0xA0, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xF0, 0xD0, 0xA0,
0x00, 0xA0, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xF0, 0xC0,
0x00, 0xA0, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xE0,

0x00, 0xA0, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xE0,
0x00, 0xA0, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xF0, 0xC0,
0x00, 0xA0, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xF0, 0xD0, 0xA0,
0x00, 0xA0, 0xF0, 0xFA,  0xFF, 0xFF, 0xFF, 0xFF,  0xFF, 0xFF, 0xFA, 0xF0,  0xE0, 0xD0, 0xB0, 0x50,

0x00, 0xA0, 0xF0, 0xF0,  0xF0, 0xF0, 0xF0, 0xF0,  0xF0, 0xF0, 0xF0, 0xE0,  0xD0, 0xB0, 0x80, 0x20,
0x00, 0x80, 0xA0, 0xA0,  0xA0, 0xA0, 0xA0, 0xA0,  0xA0, 0xA0, 0xA0, 0x90,  0x70, 0x30, 0x00, 0x00,
0x00, 0x40, 0x80, 0x80,  0x80, 0x80, 0x80, 0x80,  0x80, 0x80, 0x70, 0x50,  0x20, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00
		};

		if(gauge->orientation == GTK_ORIENTATION_VERTICAL)
		{
			gauge->tick_width = gauge->width;
			gauge->tick_height = gauge->height / nticks;
		}
		else
		{
			gauge->tick_width = gauge->width / nticks;
			gauge->tick_height = gauge->height;
		}

		g_free(gauge->tick_gray);
		if((gauge->tick_width > 0) && (gauge->tick_height > 0))
			gauge->tick_gray = (guint8 *)g_malloc(
				gauge->tick_width * gauge->tick_height * 1 * sizeof(guint8)
			);
		else
			gauge->tick_gray = NULL;
		if(gauge->tick_gray != NULL)
		{
			if(gauge->orientation == GTK_ORIENTATION_VERTICAL)
			{
				guint8 *src_tick_template_gray_rot = (guint8 *)g_malloc(
					tick_height * tick_width * 1 * sizeof(guint8)
				);
				GUIImageBufferRotateCCW90(
					1,	/* Grayscale */
					src_tick_template_gray,
					tick_width, tick_height, -1,
					src_tick_template_gray_rot,
					tick_height, tick_width, -1
				);
				GUIImageBufferResize(
					1,		/* Grayscale */
					src_tick_template_gray_rot,
					tick_height, tick_width, -1,
					gauge->tick_gray,
					gauge->tick_width, gauge->tick_height, -1,
					NULL, NULL
				);
				g_free(src_tick_template_gray_rot);
			}
			else
			{
				GUIImageBufferResize(
					1,		/* Grayscale */
					src_tick_template_gray,
					tick_width, tick_height, -1,
					gauge->tick_gray,
					gauge->tick_width, gauge->tick_height, -1,
					NULL, NULL
				);
			}
		}
	}
}

/*
 *	Draws the IFMWinGauge.
 */
void ifm_win_gauge_draw(IFMWinGauge *gauge)
{
	if(gauge == NULL)
		return;

	if(!GTK_WIDGET_VISIBLE(gauge->drawing_area))
		return;

	gauge->freeze_count++;

	switch(gauge->gauge_style)
	{
	    case IFM_WIN_GAUGE_STYLE_CONTINUOUS:
		ifm_win_gauge_draw_continuous(gauge);
		break;
	    case IFM_WIN_GAUGE_STYLE_LEDS:
		ifm_win_gauge_draw_discrete(gauge);
		break;
	    case IFM_WIN_GAUGE_STYLE_GRADIENT_LEDS:
		ifm_win_gauge_draw_led(gauge);
		break;
	}

	gauge->freeze_count--;
}

static void ifm_win_gauge_draw_continuous(IFMWinGauge *gauge)
{
	const gboolean	vertical = (gauge->orientation == GTK_ORIENTATION_VERTICAL) ?
				TRUE : FALSE;
	const gint	tar_width = gauge->width,
			tar_height = gauge->height,
			tar_bpp = 4,		/* RGBA */
			tar_bpl = tar_width * tar_bpp,
			tar_value = gauge->value * (vertical ?
				tar_height : tar_width);
	guint8 *tar_rgba = gauge->drawing_area_rgba;
	GtkWidget *w = gauge->drawing_area;
	GdkWindow *window = w->window;
	GdkDrawable *drawable = (GdkDrawable *)window;
	GtkStateType state = GTK_WIDGET_STATE(w);
	GtkStyle *style = gtk_widget_get_style(w);

	if(window == NULL)
		return;

	/* Clear the IFMWinGauge's GtkDrawingArea RGBA image buffer
	 * by drawing the background on to it
	 */
	if((tar_rgba != NULL) && (tar_width > 0) && (tar_height > 0))
	{
		GdkColor *c = &style->base[state];
		ifm_win_gauge_draw_render_base(
			tar_rgba,
			tar_width, tar_height,
			tar_bpl,
			(guint8)(c->red >> 8),
			(guint8)(c->green >> 8),
			(guint8)(c->blue >> 8)
		);
	}

	/* Render the bar to the GtkDrawingArea RGBA image buffer */
	if((tar_rgba != NULL) && (tar_width > 0) && (tar_height > 0))
	{
		GdkColor *c = &style->bg[GTK_STATE_SELECTED];
		const guint8	r = (guint8)(c->red >> 8),
				g = (guint8)(c->green >> 8),
				b = (guint8)(c->blue >> 8);
		gint x, y;
		guint8	*tar_ptr,
			*tar_line_end;

		if(vertical)
		{
			for(y = tar_height - 1; y >= 0; y--)
			{
				if((tar_height - y) > tar_value)
					break;

				for(tar_ptr = tar_rgba + (y * tar_bpl),
				    tar_line_end = tar_ptr + (tar_width * tar_bpp);
				    tar_ptr < tar_line_end;
				    tar_ptr += tar_bpp
				)
				{
					tar_ptr[0] = r;
					tar_ptr[1] = g;
					tar_ptr[2] = b;
				}
			}
		}
		else
		{
			for(x = 0; x < tar_width; x++)
			{
				if(x >= tar_value)
					break;

				for(y = 0; y < tar_height; y++)
				{
					tar_ptr = tar_rgba + (y * tar_bpl) + (x * tar_bpp);
					tar_ptr[0] = r;
					tar_ptr[1] = g;
					tar_ptr[2] = b;
				}
			}
		}
	}

	/* Send the rendered gauge image data to the IFMWinGauge's
	 * GtkDrawingArea GdkWindow
	 */
	gdk_draw_rgb_32_image(
		drawable,
		style->white_gc,
		0, 0,
		tar_width, tar_height,
		GDK_RGB_DITHER_NONE,
		(guchar *)tar_rgba,
		tar_bpl
	);
}

static void ifm_win_gauge_draw_discrete(IFMWinGauge *gauge)
{
	const gboolean	vertical = (gauge->orientation == GTK_ORIENTATION_VERTICAL) ?
				TRUE : FALSE;
	const gint	tar_width = gauge->width,
			tar_height = gauge->height,
			tar_bpp = 4,		/* RGBA */
			tar_bpl = tar_width * tar_bpp,
			tar_value = gauge->value * (vertical ?
				tar_height : tar_width);
	guint8 *tar_rgba = gauge->drawing_area_rgba;
	GtkWidget *w = gauge->drawing_area;
	GdkWindow *window = w->window;
	GdkDrawable *drawable = (GdkDrawable *)window;
	GtkStateType state = GTK_WIDGET_STATE(w);
	GtkStyle *style = gtk_widget_get_style(w);

	if(window == NULL)
		return;

	/* Clear the IFMWinGauge's GtkDrawingArea RGBA image buffer
	 * by drawing the background on to it
	 */
	if((tar_rgba != NULL) && (tar_width > 0) && (tar_height > 0))
	{
		GdkColor *c = &style->base[state];
		ifm_win_gauge_draw_render_base(
			tar_rgba,
			tar_width, tar_height,
			tar_bpl,
			(guint8)(c->red >> 8),
			(guint8)(c->green >> 8),
			(guint8)(c->blue >> 8)
		);
	}

	/* Render the ticks on to the GtkDrawingArea RGBA image buffer */
	if((tar_rgba != NULL) && (tar_width > 0) && (tar_height > 0) &&
	   (gauge->tick_gray != NULL) &&
	   (gauge->tick_width > 0) && (gauge->tick_height > 0)
	)
	{
		const gint	tick_width = gauge->tick_width,
				tick_height = gauge->tick_height,
				tick_bpp = 1,	/* Grayscale */
				tick_bpl = tick_width * tick_bpp;
		gint x, y;
		const guint8 *tick_gray = gauge->tick_gray;
		const gfloat min_gamma_coeff = 0.10f;	/* All ticks are always
						* this bright so even "off"
						* ones are still visible */
		gfloat	gamma_coeff,
			red_coeff,
			green_coeff,
			blue_coeff;
		GdkColor	*c,
				*off_c = &style->bg[state],
				*lit_c = &style->bg[GTK_STATE_SELECTED];

		if(vertical)
		{
			x = (tar_width - tick_width) / 2;
			for(y = 0; y < tar_height; y += tick_height)
			{
				if((y + tick_height) < tar_value)
				{
					gamma_coeff = 1.0f;
					c = lit_c;
				}
				else if(y >= tar_value)
				{
					gamma_coeff = min_gamma_coeff;
					c = off_c;
				}
				else
				{
					gamma_coeff = MAX(
						(gfloat)(tar_value - y) /
						(gfloat)tick_height,
						min_gamma_coeff
					);
					c = lit_c;
				}
				red_coeff = ((gfloat)c->red / (gfloat)0xFFFF) * gamma_coeff;
				green_coeff = ((gfloat)c->green / (gfloat)0xFFFF) * gamma_coeff;
				blue_coeff = ((gfloat)c->blue / (gfloat)0xFFFF) * gamma_coeff;

				ifm_win_gauge_draw_render_tick(
					gauge,
					tar_width, tar_height,
					tar_bpl,
					tar_rgba,
					tick_width, tick_height,
					tick_bpl,
					tick_gray,
					x, tar_height - y - tick_height,
					red_coeff,
					green_coeff,
					blue_coeff
				);
			}
		}
		else
		{
			y = (tar_height - tick_height) / 2;
			for(x = 0; x < tar_width; x += tick_width)
			{
				if((x + tick_width) < tar_value)
				{
					gamma_coeff = 1.0f;
					c = lit_c;
				}
				else if(x >= tar_value)
				{
					gamma_coeff = min_gamma_coeff;
					c = off_c;
				}
				else
				{
					gamma_coeff = MAX(
						(gfloat)(tar_value - x) /
						(gfloat)tick_width,
						min_gamma_coeff
					);
					c = lit_c;
				}
				red_coeff = ((gfloat)c->red / (gfloat)0xFFFF) * gamma_coeff;
				green_coeff = ((gfloat)c->green / (gfloat)0xFFFF) * gamma_coeff;
				blue_coeff = ((gfloat)c->blue / (gfloat)0xFFFF) * gamma_coeff;

				ifm_win_gauge_draw_render_tick(
					gauge,
					tar_width, tar_height,
					tar_bpl,
					tar_rgba,
					tick_width, tick_height,
					tick_bpl,
					tick_gray,
					x, y,
					red_coeff,
					green_coeff,
					blue_coeff
				);
			}
		}
	}

	/* Send the rendered gauge image data to the IFMWinGauge's
	 * GtkDrawingArea GdkWindow
	 */
	gdk_draw_rgb_32_image(
		drawable,
		style->white_gc,
		0, 0,
		tar_width, tar_height,
		GDK_RGB_DITHER_NONE,
		(guchar *)tar_rgba,
		tar_bpl
	);
}

static void ifm_win_gauge_draw_led(IFMWinGauge *gauge)
{
	const gboolean	vertical = (gauge->orientation == GTK_ORIENTATION_VERTICAL) ?
				TRUE : FALSE;
	const gint	tar_width = gauge->width,
			tar_height = gauge->height,
			tar_bpp = 4,		/* RGBA */
			tar_bpl = tar_width * tar_bpp,
			tar_value = gauge->value * (vertical ?
				tar_height : tar_width);
	guint8 *tar_rgba = gauge->drawing_area_rgba;
	GtkWidget *w = gauge->drawing_area;
	GdkWindow *window = w->window;
	GdkDrawable *drawable = (GdkDrawable *)window;
	GtkStyle *style = gtk_widget_get_style(w);

	if(window == NULL)
		return;

	/* Clear the IFMWinGauge's GtkDrawingArea RGBA image buffer by
	 * drawing the background on to it
	 */
	if((tar_rgba != NULL) && (tar_width > 0) && (tar_height > 0))
	{
		ifm_win_gauge_draw_render_base(
			tar_rgba,
			tar_width, tar_height,
			tar_bpl,
			0x00,
			0x00,
			0x00
		);
	}

	/* Render the ticks on to the GtkDrawingArea RGBA image buffer */
	if((tar_rgba != NULL) && (tar_width > 0) && (tar_height > 0) &&
	   (gauge->tick_gray != NULL) &&
	   (gauge->tick_width > 0) && (gauge->tick_height > 0)
	)
	{
		const gint	tick_width = gauge->tick_width,
				tick_height = gauge->tick_height,
				tick_bpp = 1,	/* Grayscale */
				tick_bpl = tick_width * tick_bpp;
		gint		x, y,
				xm, ym;
		gfloat color_coeff;
		const gfloat	min_gamma_coeff = 0.10f,	/* All ticks are
						* always this bright so even
						* "off" ones are still visible */
				mid_coeff = 0.68f;
		const guint8 *tick_gray = gauge->tick_gray;
		gfloat	gamma_coeff,
			red_coeff,
			green_coeff,
			blue_coeff;

		if(vertical)
		{
			x = (tar_width - tick_width) / 2;
			ym = tar_height - tick_height;
			for(y = 0; y < tar_height; y += tick_height)
			{
				if((y + tick_height) < tar_value)
					gamma_coeff = 1.0f;
				else if(y >= tar_value)
					gamma_coeff = min_gamma_coeff;
				else
					gamma_coeff = MAX(
						(gfloat)(tar_value - y) /
						(gfloat)tick_height,
						min_gamma_coeff
					);

				if(ym > 0)
					color_coeff = MIN(
						(gfloat)y / (gfloat)ym,
						1.0f
					);
				else
					color_coeff = 0.0f;
				if(color_coeff < mid_coeff)
				{
					red_coeff	= gamma_coeff * (color_coeff / mid_coeff);
					green_coeff	= gamma_coeff;
					blue_coeff	= 0.0f;
				}
				else
				{
					red_coeff	= gamma_coeff;
					green_coeff	= gamma_coeff * (1.0f -
				((color_coeff - mid_coeff) / (1.0f - mid_coeff))
							  );
					blue_coeff	= 0.0f;
				}

				ifm_win_gauge_draw_render_tick(
					gauge,
					tar_width, tar_height,
					tar_bpl,
					tar_rgba,
					tick_width, tick_height,
					tick_bpl,
					tick_gray,
					x, tar_height - y - tick_height,
					red_coeff,
					green_coeff,
					blue_coeff
				);
			}
		}
		else
		{
			y = (tar_height - tick_height) / 2;
			xm = tar_width - tick_width;
			for(x = 0; x < tar_width; x += tick_width)
			{
				if((x + tick_width) < tar_value)
					gamma_coeff = 1.0f;
				else if(x >= tar_value)
					gamma_coeff = min_gamma_coeff;
				else
					gamma_coeff = MAX(
						(gfloat)(tar_value - x) /
						(gfloat)tick_width,
						min_gamma_coeff
					);

				if(xm > 0)
					color_coeff = MIN(
						(gfloat)x / (gfloat)xm,
						1.0f
					);
				else
					color_coeff = 0.0f;
				if(color_coeff < mid_coeff)
				{
					red_coeff	= gamma_coeff * (color_coeff / mid_coeff);
					green_coeff	= gamma_coeff;
					blue_coeff	= 0.0f;
				}
				else
				{
					red_coeff	= gamma_coeff;
					green_coeff	= gamma_coeff * (1.0f -
				((color_coeff - mid_coeff) / (1.0f - mid_coeff))
							  );
					blue_coeff	= 0.0f;
				}

				ifm_win_gauge_draw_render_tick(
					gauge,
					tar_width, tar_height,
					tar_bpl,
					tar_rgba,
					tick_width, tick_height,
					tick_bpl,
					tick_gray,
					x, y,
					red_coeff,
					green_coeff,
					blue_coeff
				);
			}
		}
	}

	/* Send the rendered gauge image data to the IFMWinGauge's
	 * GtkDrawingArea GdkWindow
	 */
	gdk_draw_rgb_32_image(
		drawable,
		style->white_gc,
		0, 0,
		tar_width, tar_height,
		GDK_RGB_DITHER_NONE,
		(guchar *)tar_rgba,
		tar_bpl
	);

}

/*
 *	Renders the base.
 */
static void ifm_win_gauge_draw_render_base(
	guint8 *tar_rgba,
	const gint tar_width, const gint tar_height,
	const gint tar_bpl,
	const guint8 r,
	const guint8 g,
	const guint8 b
)
{
	const gint tar_bpp = 4;			/* RGBA */
	const guint8 a = 0xFF;
	guint8		*tar_line = tar_rgba,
			*tar_end = tar_line + (tar_height * tar_bpl),
			*tar_ptr,
			*tar_line_end;
	while(tar_line < tar_end)
	{
		tar_ptr = tar_line;
		tar_line_end = tar_ptr + tar_bpl;
		while(tar_ptr < tar_line_end)
		{
			tar_ptr[0] = r;
			tar_ptr[1] = g;
			tar_ptr[2] = b;
			tar_ptr[3] = a;
			tar_ptr += tar_bpp;
		}
		tar_line = tar_line_end;
	}
}

/*
 *	Renders the tick.
 */
static void ifm_win_gauge_draw_render_tick(
	IFMWinGauge *gauge,
	const gint tar_width, const gint tar_height,
	const gint tar_bpl,
	guint8 *tar_rgba,
	const gint src_width, const gint src_height,
	const gint src_bpl,
	const guint8 *src_gray,
	const gint x, const gint y,
	const gfloat red_coeff, const gfloat green_coeff, const gfloat blue_coeff
)
{
	const gint	tar_bpp = 4,		/* RGBA */
			src_bpp = 1;		/* Grayscale */
	gint		tx, ty,
			sx, sy;
	guint8		*tar_line,
			*tar_ptr;
	const guint8	*src_line,
			*src_ptr;

	for(sy = 0; sy < src_height; sy++)
	{
		ty = y + sy;
		if(ty < 0)
			continue;
		if(ty >= tar_height)
			break;

		tar_line = tar_rgba + (ty * tar_bpl);
		src_line = src_gray + (sy * src_bpl);

		for(sx = 0; sx < src_width; sx++)
		{
			tx = x + sx;
			if(tx < 0)
				continue;
			if(tx >= tar_width)
				break;

			tar_ptr = tar_line + (tx * tar_bpp);
			src_ptr = src_line + (sx * src_bpp);
			tar_ptr[0] = (guint8)MIN(
				(guint32)tar_ptr[0] +
					(guint32)(((guint32)*src_ptr) * red_coeff),
				(guint32)0xFF
			);
			tar_ptr[1] = (guint8)MIN(
				(guint32)tar_ptr[1] +
					(guint32)(((guint32)*src_ptr) * green_coeff),
				(guint32)0xFF
			);
			tar_ptr[2] = (guint8)MIN(
				(guint32)tar_ptr[2] +
					(guint32)(((guint32)*src_ptr) * blue_coeff),
				(guint32)0xFF
			);
		}
	}
}

/*
 *	Queues the IFMWinGauge to draw.
 */
void ifm_win_gauge_queue_draw(IFMWinGauge *gauge)
{
	if(gauge == NULL)
		return;

	gtk_widget_queue_draw(gauge->drawing_area);
}


/*
 *	Gets the IFMWinGauge's IFMWinGaugeStyle.
 */
IFMWinGaugeStyle ifm_win_gauge_get_gauge_style(IFMWinGauge *gauge)
{
	if(gauge == NULL)
		return(IFM_WIN_GAUGE_STYLE_CONTINUOUS);

	return(gauge->gauge_style);
}

/*
 *	Sets the IFMWinGauge's IFMWinGaugeStyle.
 */
void ifm_win_gauge_set_gauge_style(
	IFMWinGauge *gauge,
	const IFMWinGaugeStyle gauge_style
)
{
	if(gauge == NULL)
		return;

	if(gauge->gauge_style == gauge_style)
		return;

	gauge->gauge_style = gauge_style;
	ifm_win_gauge_queue_draw(gauge);
}


/*
 *	Sets the IFMWinGauge's value.
 */
gfloat ifm_win_gauge_get_value(IFMWinGauge *gauge)
{
	if(gauge == NULL)
		return(0.0f);

	return(gauge->value);
}

/*
 *	Sets the IFMWinGauge's value.
 */
void ifm_win_gauge_set_value(
	IFMWinGauge *gauge,
	const gfloat value
)
{
	gfloat value_clipped;

	if(gauge == NULL)
		return;

	value_clipped = CLIP(
		value,
		0.0f,
		1.0f
	);

	if(gauge->value == value_clipped)
		return;

	gauge->value = value_clipped;
	ifm_win_gauge_queue_draw(gauge);
}
