#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <paths.h>
#include <glib.h>
#include "route.h"
#include "config.h"


static const gchar *route_get_str_hex_to_uint8(
	const gchar *s,
	guint8 *v_rtn
);
static const gchar *route_get_str_hex_gsbf_to_address(
	const gchar *s,
	INAddress *v_rtn
);

static RouteItem *route_table_item_new(void);
static void route_table_item_delete(RouteItem *item);

GList *route_table_list(void);
void route_table_list_delete(GList *list);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)
#define STRPFX(s,p)     ((((s) != NULL) && ((p) != NULL)) ? \
			(!strncmp((s), (p), strlen(p))) : FALSE)
#define ISBLANK(c)		(((c) == ' ') || ((c) == '\t'))

#ifndef _PATH_PROC_NET_ROUTE
# warning "_PATH_PROC_NET_ROUTE not defined, assuming /proc/net/route"
# define _PATH_PROC_NET_ROUTE		"/proc/net/route"
#endif


/*
 *	Gets the 1 byte hex value and seeks s past the value.
 */
static const gchar *route_get_str_hex_to_uint8(
	const gchar *s,
	guint8 *v_rtn
)
{
        guint8 v[2];

	/* Seek past any "0x" prefix */
	if(STRPFX(s, "0x"))
		s += strlen("0x");

	/* Check for end of string, end of value, or end of line */
        if((*s == '\0') || ISBLANK(*s) || (*s == '\n'))
	{
		*v_rtn = 0x00;
                return(s);
	}

	/* Get the first character value */
        v[0] = (isdigit((int)*s)) ?
                (guint8)(*s - '0') : (guint8)(toupper(*s) - 'A') + 10;

	/* Seek to the next character */
        s++;

	/* Check for end of string, end of value, or end of line */
        if((*s == '\0') || ISBLANK(*s) || (*s == '\n'))
	{
		*v_rtn = 0x00;
                return(s);
	}

	/* Get the second character value */
        v[1] = (isdigit((int)*s)) ?
                (guint8)(*s - '0') : (guint8)(toupper(*s) - 'A') + 10;

	/* Set the return value */
	*v_rtn = (guint8)((v[0] * 16) + v[1]);

	/* Seek past the value and return */
	return(s + 1);
}

/*
 *	Gets the address value with greatest significant byte first and
 *	seeks s past the value.
 */
static const gchar *route_get_str_hex_gsbf_to_address(
	const gchar *s,
	INAddress *v_rtn
)
{
	gint i, n;
	const gchar *sp;

	/* Seek past any "0x" prefix */
	if(STRPFX(s, "0x"))
		s += strlen("0x");

	/* Count the number of characters in the value and divide by
	 * 2 to get the number of bytes specified by the string and
	 * not exceeding the end of the string
	 */
	for(sp = s, n = 0;
            (*sp != '\0') && !ISBLANK(*sp) && (*sp != '\n');
            sp++, n++
	);
	n = MIN((n / 2), 4);

        /* Get the address value as the greatest significant byte first
         * and seek s past the value
         */
	for(i = n - 1; i >= 0; i--)
		s = route_get_str_hex_to_uint8(
			s,
			&v_rtn->u8[i]
		);

	return(s);
}


/*
 *	Creates a new RouteItem.
 */
static RouteItem *route_table_item_new(void)
{
	return(ROUTE_ITEM(g_malloc0(sizeof(RouteItem))));
}

/*
 *	Deletes the RouteItem.
 */
static void route_table_item_delete(RouteItem *item)
{
	if(item == NULL)
		return;

	g_free(item);
}


/*
 *	Gets the routing table.
 *
 *	Returns a dynamically allocated GList of RouteItem * items.
 */
GList *route_table_list(void)
{
	const gchar	*s;
	gchar		*buf_ptr,
			*buf_end,
			buf[1024];
	GList *list;
	RouteItem *item;

	FILE *fp = fopen(_PATH_PROC_NET_ROUTE, "rb");
	if(fp == NULL)
		return(NULL);

	/*
Iface	Destination	Gateway 	Flags	RefCnt	Use	Metric	Mask		MTU	Window	IRTT
ppp0	023E9242	00000000	0005	0	0	0	FFFFFFFF	0	0	0
lo	0000007F	00000000	0001	0	0	0	000000FF	0	0	0
ppp0	00000000	023E9242	0003	0	0	0	00000000	0	0	0
	 */

	list = NULL;
	while(fgets(buf, sizeof(buf), fp) != NULL)
	{
		s = buf;
		while(ISBLANK(*s))
			s++;

		/* Skip the header */
		if(STRPFX(s, "Iface"))
			continue;

		item = route_table_item_new();
		if(item == NULL)
			break;

#define SEEK_PAST_SPACES(_s_)	{			\
 while(ISBLANK(*(_s_)))					\
  (_s_)++;						\
}
#define SEEK_PAST_WORD(_s_)	{			\
 while((*s != '\0') && !ISBLANK(*(_s_)))		\
  (_s_)++;						\
}

		/* Interface Name */
		buf_ptr = item->iface;
		buf_end = item->iface + sizeof(item->iface);
		while((*s != '\0') && !ISBLANK(*s) && (*s != '\n') &&
		      (buf_ptr < buf_end)
		)
		{
			*buf_ptr = *s;
			buf_ptr++;
			s++;
		}
		if(buf_ptr < buf_end)
			*buf_ptr = '\0';
		else
			*(buf_end - 1) = '\0';
		SEEK_PAST_SPACES(s);

		/* Destination Address */
		s = route_get_str_hex_gsbf_to_address(
			s,
			&item->destination
		);
		SEEK_PAST_SPACES(s);

		/* Gateway Address */
		s = route_get_str_hex_gsbf_to_address(
			s,
			&item->gateway
		);
		SEEK_PAST_SPACES(s);

		/* Flags */
		item->flags = (RouteFlags)atoi((const char *)s);
		SEEK_PAST_WORD(s);
		SEEK_PAST_SPACES(s);

		/* Reference Count */
		item->refcnt = (gulong)atol((const char *)s);
                SEEK_PAST_WORD(s);
                SEEK_PAST_SPACES(s);

		/* Use */
		item->use = (gulong)atol((const char *)s);
                SEEK_PAST_WORD(s);
                SEEK_PAST_SPACES(s);

		/* Metric */
		item->metric = (gulong)atol((const char *)s);
                SEEK_PAST_WORD(s);
                SEEK_PAST_SPACES(s);

		/* Mask */
		s = route_get_str_hex_gsbf_to_address(
			s,
			&item->mask
		);
		SEEK_PAST_SPACES(s);

		/* MTU */
		item->mtu = (gulong)atol((const char *)s);
                SEEK_PAST_WORD(s);
                SEEK_PAST_SPACES(s);

		/* Window */
		item->window = (gulong)atol((const char *)s);
                SEEK_PAST_WORD(s);
                SEEK_PAST_SPACES(s);

		/* IRTT */
		item->irtt = (gulong)atol((const char *)s);
                SEEK_PAST_WORD(s);
                SEEK_PAST_SPACES(s);

#undef SEEK_PAST_WORD
#undef SEEK_PAST_SPACES

		list = g_list_append(
			list,
			item
		);
	}

	(void)fclose(fp);

	return(list);
}

/*
 *	Deletes the route table list.
 *
 *	The list specifies the GList of RouteItem * items.
 */
void route_table_list_delete(GList *list)
{
	if(list == NULL)
		return;

	g_list_foreach(list, (GFunc)route_table_item_delete, NULL);
	g_list_free(list);
}
